package device

import (
	"crypto/rand"
	"encoding/hex"
	"net"
	"net/http"

	"git.autistici.org/id/auth"
	"github.com/gorilla/sessions"
	"github.com/mssola/user_agent"
	"github.com/oschwald/maxminddb-golang"
)

func randomDeviceID() string {
	b := make([]byte, 8)
	rand.Read(b)
	return hex.EncodeToString(b)
}

// Manager can provide DeviceInfo entries for incoming HTTP requests.
type Manager struct {
	store             sessions.Store
	geodb             *maxminddb.Reader
	trustedForwarders []net.IPNet
	remoteAddrHeader  string
}

// Config stores options for the device info manager.
type Config struct {
	AuthKey           []byte   `yaml:"auth_key"`
	GeoIPDataFile     string   `yaml:"geo_ip_data"`
	TrustedForwarders []string `yaml:"trusted_forwarders"`
	RemoteAddrHeader  string   `yaml:"remote_addr_header"`
}

// New returns a new Manager with the given configuration.
func New(config *Config) (*Manager, error) {
	if config == nil {
		config = &Config{}
	}
	tf, err := parseIPNetList(config.TrustedForwarders)
	if err != nil {
		return nil, err
	}

	var geodb *maxminddb.Reader
	if config.GeoIPDataFile != "" {
		geodb, err = maxminddb.Open(config.GeoIPDataFile)
		if err != nil {
			return nil, err
		}
	}

	// The remote IP header (if any) defaults to X-Forwarded-For.
	hdr := "X-Forwarded-For"
	if config.RemoteAddrHeader != "" {
		hdr = config.RemoteAddrHeader
	}

	return &Manager{
		geodb:             geodb,
		store:             newStore(config.AuthKey),
		trustedForwarders: tf,
		remoteAddrHeader:  hdr,
	}, nil
}

const deviceIDSessionName = "_dev"

// GetDeviceInfoFromRequest will retrieve or create a DeviceInfo
// object for the given request. It will always return a valid object.
// The ResponseWriter is needed to store the unique ID on the client
// when a new device info object is created.
func (m *Manager) GetDeviceInfoFromRequest(w http.ResponseWriter, req *http.Request) *auth.DeviceInfo {
	session, _ := m.store.Get(req, deviceIDSessionName)
	devID, ok := session.Values["id"].(string)
	if !ok || devID == "" {
		// Generate a new Device ID and save it on the client.
		devID = randomDeviceID()
		session.Values["id"] = devID
		session.Save(req, w)
	}

	uaStr := req.UserAgent()
	ua := user_agent.New(uaStr)
	browser, _ := ua.Browser()
	d := auth.DeviceInfo{
		ID:        devID,
		UserAgent: uaStr,
		Mobile:    ua.Mobile(),
		OS:        ua.OS(),
		Browser:   browser,
	}

	if ip := m.getIPFromRequest(req); ip != nil {
		d.RemoteAddr = ip.String()
		if zone, err := m.getZoneForIP(ip); err == nil {
			d.RemoteZone = zone
		}
	}

	return &d
}
